import time

import dateutil
tzinfos = {"GMT": dateutil.tz.gettz("GMT")}
import numpy as np

from crossnetwork_processing_functions import get_data_files, create_GHOST_stage_logfile
from GHOST_standards import standard_parameters, get_standard_data, get_standard_metadata, standard_data_flag_name_to_data_flag_code, standard_QA_name_to_QA_code
from network_file_standardise import NetworkFileStandardise
from network_parsers import ParseAERONETFile
from standardisation_mapping_functions import map_network_measurement_information_to_standard_measurement_information, network_data_flag_to_standard_data_flag_mapping
from time_functions import round_to_minute, datetime_to_fractional_minutes_from_reference, get_datafile_download_timestamp


class StandardiseNetworkData:

    '''Convert daily provider original files in a month to GHOST standard numpy station specific files, per data level.'''

    def __init__(self, month, run_details, parameter_details, network_details):
        
        #add passed variables to self
        self.month = month
        self.run_details = run_details
        self.parameter_details = parameter_details
        self.network_details = network_details


    def do_work(self):
        '''Function which does executes the functions of the class in a pipeline'''

        #gather data files per data level to parse
        lev10_data_files, lev15_data_files, lev20_data_files = self.gather_data_files()

        #parse all data files per data level (writing standard files out per station)
        for data_level, data_files in zip(['1.0','1.5','2.0'], [lev10_data_files, lev15_data_files, lev20_data_files]):
            self.parse_files(data_level, data_files)

        #write a valid process log file
        self.run_details['log_file_str'] += str((time.time() - self.run_details['stage_process_start_time'])/60.)
        create_GHOST_stage_logfile(0,self.run_details)


    def gather_data_files(self):

        '''Function to gather data files to read per data level'''

        #get specific data files for current month
        data_files = get_data_files(self.run_details, self.parameter_details, specific_month=self.month)
    
        #define select variable fields (also define date/time variable names)
        if self.parameter_details['aeronet_matrix'] == 'oneill':
            self.date_var = 'Date_(dd:mm:yyyy)'
            self.time_var = 'Time_(hh:mm:ss)'
            self.req_vars = ['AERONET_Site',self.date_var,self.time_var,self.parameter_details['aeronet_parameter_name'],'Fine_Mode_AOD_500nm[tau_f]','Data_Quality_Level','AERONET_Instrument_Number','AERONET_Site_Name','Site_Latitude(Degrees)','Site_Longitude(Degrees)','Site_Elevation(m)','contact_name','contact_institution','contact_email_address']
        elif parameter_details['aeronet_matrix'] == 'directsun':   
            self.date_var = 'Date(dd:mm:yyyy)'
            self.time_var = 'Time(hh:mm:ss)'
            if self.run_details['parameter'] == 'AOD_550nm':
                self.req_vars = ['AERONET_Site',self.date_var,self.time_var,'AOD_440nm','AOD_675nm','AOD_870nm','440-870_Angstrom_Exponent','Data_Quality_Level','AERONET_Instrument_Number','AERONET_Site_Name','Site_Latitude(Degrees)','Site_Longitude(Degrees)','Site_Elevation(m)','contact_name','contact_institution','contact_email_address']
            else:
                self.req_vars = ['AERONET_Site',self.date_var,self.time_var,self.parameter_details['aeronet_parameter_name'],'Data_Quality_Level','AERONET_Instrument_Number','AERONET_Site_Name','Site_Latitude(Degrees)','Site_Longitude(Degrees)','Site_Elevation(m)','contact_name','contact_institution','contact_email_address']

        #split data files into different lists by data level
        lev10_data_files = []
        lev15_data_files = []
        lev20_data_files = []
        for data_file in data_files:
            if '/lev1.0/' in data_file:
                lev10_data_files.append(data_file)
            elif '/lev1.5/' in data_file:
                lev15_data_files.append(data_file)
            elif '/lev2.0/' in data_file:
                lev20_data_files.append(data_file)

        return lev10_data_files, lev15_data_files, lev20_data_files


    def parse_files(self, data_level, data_files):

        '''Function to parse data files in a month for a data level to GHOST standard numpy station specific files'''
        
        self.run_details['log_file_str'] += 'PROCESSING DATA LEVEL: {}\n'.format(data_level) 

        #define dictionary for storing all parsed data (per specific data level)
        all_data = {k:[] for k in self.req_vars}
    
        #iterate through daily data files (for specific data level)
        for data_file in data_files:
            self.run_details['log_file_str'] += 'PROCESSING FILE: {}\n'.format(data_file) 

            #create object for parsing AERONET daily file
            parsing_obj = ParseAERONETFile()

            #open and read file contents into memory
            f = open(data_file, "r", encoding='latin1')
            page = f.read()
            f.close()

            #parse read file contents 
            parsing_obj.feed(page)

            #iterate through read object dictionary keys
            for var in list(parsing_obj.var_data.keys()):
                #is read object key one of required vars?
                if var in self.req_vars:
                    #if so --> append to all data dictionary
                    all_data[var] = np.append(all_data[var], np.array(parsing_obj.var_data[var], dtype=np.object))

            #add parsed contact information to all_data dictionary
            all_data['contact_name'] = np.append(all_data['contact_name'], np.array([parsing_obj.contact_name]*len(parsing_obj.var_data['AERONET_Site']), dtype=np.object))
            all_data['contact_institution'] = np.append(all_data['contact_institution'], np.array([parsing_obj.contact_institution]*len(parsing_obj.var_data['AERONET_Site']), dtype=np.object)) 
            all_data['contact_email_address'] = np.append(all_data['contact_email_address'], np.array([parsing_obj.contact_email_address]*len(parsing_obj.var_data['AERONET_Site']), dtype=np.object)) 
            
        #get all unique "AERONET_Site" variables --> take these to be the unique station references
        unique_station_references = np.unique(all_data['AERONET_Site'])

        #get current AERONET data level in format as reported in file
        data_level_mod = 'lev'+data_level.replace(".","") 
        #get mask of all data of correct data level (for sanity)
        data_level_bool = np.array([True if var == data_level_mod else False for var in all_data['Data_Quality_Level']])

        #get dictionary necessary for mapping of AERONET file data flags to standard data flag codes
        data_flag_mapping_dictionary = network_data_flag_to_standard_data_flag_mapping['AERONET']

        #iterate through unique stations references in data
        for orig_station_reference in unique_station_references:

            start_timer = time.time()

            #set flag to know when have written data for station 
            written_data = False

            #strip underscore from original unique station reference string
            station_reference = orig_station_reference.replace('_','')

            self.run_details['log_file_str'] += '{} -- PARSING RAW DATA\n'.format(station_reference)

            #create site instances of metadata/data dictionaries
            site_data =     get_standard_data(self.parameter_details)
            site_metadata = get_standard_metadata(self.parameter_details)

            #isolate variable data specific for station
            unique_station_bool = all_data['AERONET_Site'] == orig_station_reference
            #further isolate variable data to be only that associated with a data quality level that is currently being processed 
            unique_station_bool = data_level_bool & unique_station_bool
            station_var_data = {req_var:all_data[req_var][unique_station_bool] for req_var in self.req_vars}
            if self.run_details['parameter'] == 'AOD_550nm':
                station_var_data['AOD_550nm'] = []

            #------------------------------------------------------------------------#
            #PARSE TIME/DATA
            #------------------------------------------------------------------------#
        
            #parse datetime for each reported timepoint 
            #first create a unified date/time string (yyyy-mm-dd hh:mm:ss Z, of the measurement --> z is always GMT)
            #secondly parse into datetime (taking into account timezone information). then convert to UTC datetime, then round to nearest minute    
            start_dt = [round_to_minute(dateutil.parser.parse('{}-{}-{} {} GMT'.format(date[6:],date[3:5],date[:2],time), tzinfos=tzinfos).astimezone(dateutil.tz.tzutc())).replace(tzinfo=None) for date,time in zip(station_var_data[self.date_var],station_var_data[self.time_var])] 
            #convert each start measurement datetime to minutes from reference time
            frac_time = [datetime_to_fractional_minutes_from_reference(dt) for dt in start_dt]

            #------------------------------------------------------------------------#
            #PARSE METADATA
            #------------------------------------------------------------------------#

            #GHOST version
            site_metadata['GHOST_version']['value'] = [self.run_details['ghost_version']]*len(frac_time)

            #initialise string which will store all GHOST process warnings as empty            
            site_metadata['process_warnings']['value'] = ['']*len(frac_time)

            #check if measurement position metadata is suspect and write it to metadata/data dictionaries
            for meta_ii in range(len(frac_time)):
                site_metadata['longitude']['value'].append(station_var_data['Site_Longitude(Degrees)'][meta_ii])
                site_metadata['latitude']['value'].append(station_var_data['Site_Latitude(Degrees)'][meta_ii])
                site_metadata['altitude']['value'].append(station_var_data['Site_Elevation(m)'][meta_ii])

            #save meta_update_stamp, set as copy of time array
            site_metadata['meta_update_stamp']['value'] = frac_time

            #write known metadata  
            site_metadata['network']['value'] = [self.run_details['network']]*len(frac_time)
            site_metadata['station_name']['value'] = station_var_data['AERONET_Site_Name']
            site_metadata['data_level']['value'] = [data_level]*len(frac_time)
            site_metadata['contact_name']['value'] = station_var_data['contact_name']
            site_metadata['contact_institution']['value'] = station_var_data['contact_institution']
            site_metadata['contact_email_address']['value'] = station_var_data['contact_email_address']
            site_metadata['principal_investigator_name']['value'] = station_var_data['contact_name']
            site_metadata['principal_investigator_institution']['value'] = station_var_data['contact_institution']
            site_metadata['principal_investigator_email_address']['value'] = station_var_data['contact_email_address']

            #process measurement information
            standard_measurement_metadata, measurement_qa_flags, standard_method_abbreviation = map_network_measurement_information_to_standard_measurement_information(self.parameter_details['aeronet_matrix'], site_metadata, self.run_details, self.parameter_details, write_to_metadata_array=False)
            for key, value in standard_measurement_metadata.items():
                site_metadata[key]['value'] = [value] * len(frac_time)
            site_metadata['measurement_qa_flags'] = {'value':[], 'data_type':np.object, 'string_format':'lower_short'}
            measurement_qa_flag_codes = []
            #set measurement qa codes which are True as joined metadata str
            for standard_flag_name in measurement_qa_flags.keys():
                if measurement_qa_flags[standard_flag_name]:
                    measurement_qa_flag_codes.append(str(standard_QA_name_to_QA_code[standard_flag_name]))
            site_metadata['measurement_qa_flags']['value'] = [','.join(measurement_qa_flag_codes)]*len(frac_time)

            #append station_reference now that we have the standard method abbreviation
            site_metadata['station_reference']['value'] = ['{}_{}'.format(station_reference, standard_method_abbreviation)]*len(frac_time)

            #iterate through metadata variables, if any columns have not been appended to add a NaN
            for column_header in list(site_metadata.keys()):
                n_column = len(site_metadata[column_header]['value'])
                #if column is empty, then append NaN
                if n_column < 1:
                    site_metadata[column_header]['value'] = [np.NaN] * len(frac_time)

            #get data file download timestamp
            site_metadata['data_download_stamp']['value'] = [get_datafile_download_timestamp(data_file)] * len(frac_time)

            #------------------------------------------------------------------------#
            #PARSE FLAGS/QA
            #------------------------------------------------------------------------#

            #create empty list for standard data flags associated with each measurement
            flags = []
            qa = []
 
            #process data flags (artificially created)
            for val_ii, val in enumerate(station_var_data[self.date_var]):

                #lists to append qa/data flags per measurement
                flags_per_measurement = []
                qa_per_measurement = []

                #set GHOST/network decreed validity as initially 'V'
                ghost_decreed_validity = 'V'
                network_decreed_validity = 'V'
 
                #for AERONET O'Neill matrix measurements, artificially create QA flag variable where FINE_MODE_FRAC_500nm is outside valid bounds
                if self.parameter_details['aeronet_matrix'] == 'oneill':
 
                    #is FINE_MODE_FRAC_500nm outside valid bounds?
                    float_val = float(station_var_data['Fine_Mode_AOD_500nm[tau_f]'][val_ii])
                    if (float_val < standard_parameters['FINE_MODE_FRAC_500nm']['extreme_lower_limit']) or (float_val > standard_parameters['FINE_MODE_FRAC_500nm']['extreme_upper_limit']):
                        data_flag = 'FM500FRAC_I' 
                        # map data flag/s and append
                        standard_data_flag_dict = data_flag_mapping_dictionary[data_flag]
                        #get mapped standard data flags
                        standard_data_flag_names = standard_data_flag_dict['standard_data_flag_name']
                        #iterate through standard data flags 
                        for standard_data_flag_ii, standard_data_flag_name in enumerate(standard_data_flag_names):
                            #check if standard data flag is empty (i.e. unmapped as non-useful)
                            if standard_data_flag_name != '': 
                                #get corresponding data flag code for standard data flag
                                flag_code = standard_data_flag_name_to_data_flag_code[standard_data_flag_name]
                                #append standard data flag
                                flags_per_measurement.append(flag_code)
                                #determine if GHOST/network decreed valid
                                if standard_data_flag_dict['GHOST_decreed_validity'][standard_data_flag_ii] == 'I':
                                    ghost_decreed_validity = 'I'
                                if standard_data_flag_dict['network_decreed_validity'][standard_data_flag_ii] == 'I':
                                    network_decreed_validity = 'I'

                #when calculating AOD 550nm from AERONET directsun measurements, artificially create QA flag variable where one of used components is outside valid bounds
                elif (self.parameter_details['aeronet_matrix'] == 'directsun') & (self.run_details['parameter'] == 'AOD_550nm'): 
                
                    #Is AOD_440nm, AOD675nm, AOD870nm or AE_440-870nm outside valid bounds?
                    #calculate AOD_550nm (if all needed component values in valid ranges)
                    float_aod440 = float(station_var_data['AOD_440nm'][val_ii]) 
                    float_aod675 = float(station_var_data['AOD_675nm'][val_ii]) 
                    float_aod870 = float(station_var_data['AOD_870nm'][val_ii]) 
                    float_ae440_870 = float(station_var_data['440-870_Angstrom_Exponent'][val_ii]) 
                    aod440_valid = False
                    aod675_valid = False
                    aod870_valid = False
                    ae440_870_valid = False

                    #AE_440-870nm outside valid bounds?
                    if (float_ae440_870 >= standard_parameters['AE_440-870nm']['extreme_lower_limit']) & (float_ae440_870 <= standard_parameters['AE_440-870nm']['extreme_upper_limit']):
                        ae440_870_valid = True
                        #AOD_440nm outside valid bounds?
                        if (float_aod440 >= standard_parameters['AOD_440nm']['extreme_lower_limit']) & (float_aod440 <= standard_parameters['AOD_440nm']['extreme_upper_limit']):
                            aod_550_adjust1 = (float_aod440/(440**(-float_ae440_870)))*(550**(-float_ae440_870))
                            aod440_valid = True
                        #AOD_675nm outside valid bounds?
                        if (float_aod675 >= standard_parameters['AOD_675nm']['extreme_lower_limit']) & (float_aod675 <= standard_parameters['AOD_675nm']['extreme_upper_limit']):
                            aod_550_adjust2 = (float_aod675/(675**(-float_ae440_870)))*(550**(-float_ae440_870))
                            aod675_valid = True
                        #AOD_870nm outside valid bounds?
                        if (float_aod870 >= standard_parameters['AOD_870nm']['extreme_lower_limit']) & (float_aod870 <= standard_parameters['AOD_870nm']['extreme_upper_limit']):   
                            aod_550_adjust3 = (float_aod870/(870**(-float_ae440_870)))*(550**(-float_ae440_870))
                            aod870_valid = True

                    #if one of needed values is outside of valid bounds then append data flag, and also set AOD 550nm value as NaN
                    if (not aod440_valid) or (not aod675_valid) or (not aod870_valid) or (not ae440_870_valid):
                        data_flag = 'AOD550CON_I' 
                        # map data flag/s and append
                        standard_data_flag_dict = data_flag_mapping_dictionary[data_flag]
                        #get mapped standard data flags
                        standard_data_flag_names = standard_data_flag_dict['standard_data_flag_name']
                        #iterate through standard data flags 
                        for standard_data_flag_ii, standard_data_flag_name in enumerate(standard_data_flag_names):
                            #check if standard data flag is empty (i.e. unmapped as non-useful)
                            if standard_data_flag_name != '': 
                                #get corresponding data flag code for standard data flag
                                flag_code = standard_data_flag_name_to_data_flag_code[standard_data_flag_name]
                                #append standard data flag
                                flags_per_measurement.append(flag_code)
                                #determine if GHOST/network decreed valid
                                if standard_data_flag_dict['GHOST_decreed_validity'][standard_data_flag_ii] == 'I':
                                    ghost_decreed_validity = 'I'
                                if standard_data_flag_dict['network_decreed_validity'][standard_data_flag_ii] == 'I':
                                    network_decreed_validity = 'I'
                        station_var_data['AOD_550nm'].append(np.NaN)
                    #append valid calculated AOD 500nm
                    else:
                        station_var_data['AOD_550nm'].append(np.mean([aod_550_adjust1,aod_550_adjust2,aod_550_adjust3]))
 
                #if ghost decreed validity, or network decreed validity is Invalid, then set QA flags
                if ghost_decreed_validity == 'I':
                    qa_per_measurement.append(standard_QA_name_to_QA_code['Invalid Data Provider Flags - GHOST Decreed'])
                if network_decreed_validity == 'I':
                    qa_per_measurement.append(standard_QA_name_to_QA_code['Invalid Data Provider Flags - Network Decreed'])           

                #append flags/qa codes to list to write out
                flags.append(flags_per_measurement)
                qa.append(qa_per_measurement)

            #----------------------------------------------------------------------------#
            #WRITE OUT DATA
            #----------------------------------------------------------------------------#
            site_data['time']['value'] = frac_time
            site_data[self.parameter_details['bsc_parameter_name']]['value'] = station_var_data[self.parameter_details['aeronet_parameter_name']]
            site_data['temporal_resolution']['value'] = [0]*len(frac_time)   
            site_data['flag']['value'] = flags
            site_data['qa']['value'] = qa
            #iterate through data variables, if any columns have not been appended to, add a list of NaNs (equal to length of time array)
            for column_header in list(site_data.keys()):
                n_column = len(site_data[column_header]['value'])
                #if column is empty, then append NaN
                if n_column == 0:
                     if site_data[column_header]['data_type'] != np.uint8:
                         site_data[column_header]['value'] = [np.NaN]*len(site_data['time']['value'])
                     else:
                         site_data[column_header]['value'] = [255]*len(site_data['time']['value'])
            #------------------------------------------------------
 
            #standardise metadata/data formats and write out station data to numpy file
            NetworkFileStandardise(site_metadata, site_data, self.run_details, self.parameter_details)