import bisect
import copy
import datetime
import glob
import os
import re
import shutil
import sys
import time

import numpy as np
import pandas as pd

from GHOST_standards import standard_data_flag_name_to_data_flag_code, standard_QA_name_to_QA_code


def get_configuration_variables(slurm_job_ID, network, parameter):

    '''return relevant specific dictionaries for specific network/parameter configuration'''

    from GHOST_standards import standard_parameters, standard_networks
    import importlib

    #Read configuration file (as was originally set when pipeline was launched)
    cwd = os.getcwd()
    if '/run' in cwd:
        sys.path.append(cwd.split('/run')[0]+'/run/GHOST_configurations')
    elif '/processing_modules' in cwd:
        sys.path.append(cwd.split('/processing_modules')[0]+'/run/GHOST_configurations')
    GHOST_configuration = importlib.import_module('configuration_{}'.format(slurm_job_ID))

    #get run details
    start_year =  GHOST_configuration.start_date[:4]
    end_year =    GHOST_configuration.end_date[:4]
    start_month = GHOST_configuration.start_date[4:]
    end_month =   GHOST_configuration.end_date[4:]
    run_details = {k: v for k, v in vars(GHOST_configuration).items() if not k.startswith('_')}
    run_details['slurm_job_ID'] = slurm_job_ID
    run_details['network'] = network
    run_details['parameter'] = parameter
    run_details['start_year'] = start_year
    run_details['end_year'] = end_year
    run_details['start_month'] = start_month
    run_details['end_month'] = end_month
    
    #get standard parameter details
    parameter_details = standard_parameters[parameter]
        
    #get standard network details
    network_details = standard_networks[network]

    return run_details, parameter_details, network_details


def create_GHOST_stage_logfile(process_code, run_details, previous_stage_data_files=[]):
    
    '''function which creates a logfile for stating outcome of a GHOST stage.
       the filename is prefixed with a code referencing the stage outcome + param_to_parallelise_by: process_code_param_to_parallelise_by.out
       the process codes are:
       0: Process completed without issue
       1: Process completed, but station no longer continuing in pipeline
       2: Caught error in process
       3: Uncaught error in process

       remove previous stage temporary data files associated with param_to_parallelise_by if configured to, if not stage S1/T1, and process code is 0 or 1
    '''
    f=open('{}/run/stage_logs/{}_{}/{}/{}_{}.out'.format(run_details['process_root'],run_details['network'],run_details['parameter'],run_details['GHOST_stage'],process_code,run_details['log_file_ref']),"w")
    f.write(str(run_details['log_file_str']))
    f.close() 

    #delete previous stage temporary data files associated with param_to_parallelise_by if configured to, if not stage S1/T1, and process code is 0 or 1
    if run_details['remove_temporary_data_files_after_stage']:
        if run_details['GHOST_stage'] not in ['S1','T1']:
            if int(process_code) in [0,1]:
                if type(previous_stage_data_files) == list: 
                    for data_file in previous_stage_data_files:
                        os.remove(data_file)
                else:
                    os.remove(previous_stage_data_files)

    #exit from current process after writing logfile
    sys.exit()
    

def check_directory_existence(directory_tree_str, directories_not_to_test=None):

    '''iterate through provided directory tree string
       first, check if directory exists, if not create it.
       second, check if each directory has 770 permissions.
       finally, also ensure group owner is bsc32.
    '''

    #define special characters that will need to be escaped for execution of commands      
    special_characters = ['(',')']

    #first check if instructed to not to check existence of part of directory tree str 
    if directories_not_to_test is not None:
        #modify directory_tree_str to not include directories_not_to_test
        directory_tree_str = directory_tree_str.replace(directories_not_to_test,'')
        #set directory_str_to_test to be directories_not_to_test
        directory_str_to_test = copy.deepcopy(directories_not_to_test)
    else:
        #else, set directory_str_to_test as initially empty string
        directory_str_to_test = ''

    #split directory tree str
    directory_tree_str_split = directory_tree_str.split('/')

    #iterate through directory_tree_str_split
    for current_directory in directory_tree_str_split:

        #if current directory is empty string then do not check directory existence
        if current_directory != '':

            #add current_dictionary to directory_str_to_test
            directory_str_to_test = directory_str_to_test + '/' + current_directory

            #does directory_str_to_test exist? 
            if not os.path.isdir(directory_str_to_test):
                
                #escape certain special characters in directory_str_to_test
                alt_directory_str_to_test = copy.deepcopy(directory_str_to_test)
                for ch in special_characters:
                    alt_directory_str_to_test = alt_directory_str_to_test.replace(ch,'\{}'.format(ch))

                #if not, create it
                os.system("mkdir {}".format(alt_directory_str_to_test))
                #give 770 permissions to directory
                os.system("chmod 770 {}".format(alt_directory_str_to_test))
                #make group owner bsc32
                os.system("chgrp bsc32 {}".format(alt_directory_str_to_test))


def set_file_permissions_ownership(file_str):
    '''set 770 permissions for a newly written file and also ensure group owner is bsc32'''

    #define special characters that will need to be escaped for execution of commands      
    special_characters = ['(',')']

    #escape certain special characters in file_str
    for ch in special_characters:
        file_str = file_str.replace(ch,'\{}'.format(ch))

    #give 770 permissions to file
    os.system("chmod 770 {}".format(file_str))

    #make group owner bsc32
    os.system("chgrp bsc32 {}".format(file_str))
        

def check_GHOST_output_directories(run_details, parameter_details, stage_type='S'):

    '''test for presence of native / output resolution directories for defined stage type, for a specific network/parameter.
       if directories do not exist, create them.
    '''

    from GHOST_standards import standard_temporal_resolutions

    #S stage type --> then create native directory 
    if stage_type == 'S':
        
        #set native directory for specific network/parameter
        native_dir = run_details['data_root'] + '{}/{}/native/{}/'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'])
    
        #check if need to create any directories in path 
        check_directory_existence(native_dir, directories_not_to_test=run_details['data_root'])

    #T stage type --> then create output resolution directories
    elif stage_type == 'T':
        #iterate through all defined temporal resolutions to process
        for temporal_resolution_code in run_details['temporal_resolutions']:

            #get temporal resolution specific variables
            temporal_resolution_vars = standard_temporal_resolutions[temporal_resolution_code]

            #iterate through all defined regions to process
            for region in run_details['regions']:

                #set network directory per region
                if region == 'GLOBAL':
                    network_dir = run_details['network']
                else:
                    network_dir = '{}_{}'.format(run_details['network'], region)

                #set output resolution directory for specific network/output resolution/parameter
                output_resolution_dir = run_details['data_root'] + '{}/{}/{}/{}/'.format(network_dir,run_details['ghost_version'],temporal_resolution_vars['temporal_resolution_path'],parameter_details['bsc_parameter_name'])
        
                #check if need to create any directories in path 
                check_directory_existence(output_resolution_dir, directories_not_to_test='{}{}/{}'.format(run_details['data_root'],network_dir,run_details['ghost_version']))
    

def check_GHOST_stage_temporary_directories(run_details, parameter_details, stage='S1'):

    '''test for presence of a specific stages'/all temporary data file directories for a specific network/parameter,
       used to store .npz station files between each of the S stages,
       and year/month chunked station files between the T stages. 
       if directories do not exist, create them.
       if directories do exist, delete all current temporary data files in directories
    '''

    from GHOST_standards import standard_temporal_resolutions

    S_GHOST_stages = []
    #if 'all' are defined to be checked get all defined S stages to process
    if stage == 'all':
        S_GHOST_stages = sorted([stage for stage in run_details['stages_to_process'] if 'S' in stage])
        #should not have a temporary directory for S3/S4
        if 'S3' in S_GHOST_stages:
            S_GHOST_stages.remove('S3')
            S3_dir = run_details['data_root']+'{}/{}/native/{}/temporary_S3/'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'])
            if os.path.isdir(S3_dir):
                shutil.rmtree(S3_dir) 
        if 'S4' in S_GHOST_stages:
            S_GHOST_stages.remove('S4')
            S4_dir = run_details['data_root']+'{}/{}/native/{}/temporary_S4/'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'])
            if os.path.isdir(S4_dir):
                shutil.rmtree(S4_dir) 

    #elif specific S stage to be checked?
    elif 'S' in stage:
        S_GHOST_stages = [stage]      

    #iterate through all defined S GHOST stage directories to check 
    for S_stage in S_GHOST_stages:
        #define temporary data directory for GHOST stage    
        temp_dir = run_details['data_root']+'{}/{}/native/{}/temporary_{}/'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'],S_stage)
        
        #check if need to create any directories in path 
        check_directory_existence(temp_dir, directories_not_to_test='{}{}/{}/native/{}'.format(run_details['data_root'],run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name']))
        
        #remove all files currently in the directory
        for zippath in glob.iglob(os.path.join(temp_dir, '*.npz')):
            os.remove(zippath)    

    #check if 'T1'/'all' is defined
    if stage in ['T1','all']:
        #if so, check year/month temporary directories for all defined temporal resolutions to output
        
        #iterate through defined temporal_resolutions to output
        for temporal_resolution_code in run_details['temporal_resolutions']:

            #iterate through defined regions to process
            for region in run_details['regions']:

                #set network directory per region
                if region == 'GLOBAL':
                    network_dir = run_details['network']
                else:
                    network_dir = '{}_{}'.format(run_details['network'], region)

                #get all years to process
                process_years = range(int(run_details['start_year']),int(run_details['end_year'])+1)
                for year in process_years:

                    #is it first process year?
                    if year == process_years[0]:
                        months = np.arange(int(run_details['start_month']), 13)
    
                    #is it last process year?
                    elif year == process_years[-1]:
                        months = np.arange(1,int(run_details['end_month']))
    
                    #full year of months otherwise
                    else:
                        months = np.arange(1,13)
    
                    for month in months:
                        if month < 10:
                            month_string = '0'+str(month)
                        else:
                            month_string = str(month)
                        #define temporary data directory for T1 stage  
                        temp_dir = run_details['data_root']+'{}/{}/{}/{}/temporary/{}/{}/'.format(network_dir,run_details['ghost_version'],standard_temporal_resolutions[temporal_resolution_code]['temporal_resolution_path'],parameter_details['bsc_parameter_name'],year,month_string)

                        #check if need to create any directories in path 
                        check_directory_existence(temp_dir, directories_not_to_test='{}{}/{}/{}/{}'.format(run_details['data_root'],network_dir,run_details['ghost_version'],standard_temporal_resolutions[temporal_resolution_code]['temporal_resolution_path'],parameter_details['bsc_parameter_name']))
    
                        #remove all files currently in the directory
                        for zippath in glob.iglob(os.path.join(temp_dir, '*.npz')):
                            os.remove(zippath)
                        for zippath in glob.iglob(os.path.join(temp_dir, '*.nc')):
                            os.remove(zippath)


def check_GHOST_stage_arguments_logs_submit(run_details):

    '''clean all existing argument/log/submit directories for defined GHOST stages to process.
       if directories do not already exist, create them.'''

    #iterate through defined GHOST stages to process
    for GHOST_stage in run_details['stages_to_process']:

        #define directory where submittal logs will be written
        log_dir = run_details['process_root']+'/run/stage_logs/{}_{}/{}'.format(run_details['network'],run_details['parameter'], GHOST_stage)

        #check if need to create any directories in path
        check_directory_existence(log_dir, directories_not_to_test=run_details['process_root'])
    
        #clean log directory
        for zippath in glob.iglob(os.path.join(log_dir, '*.out')):
            os.remove(zippath)
        for zippath in glob.iglob(os.path.join(log_dir, '*.err')):
            os.remove(zippath)

        #-----------------------------#

        #define directory where job arguments will be written
        arguments_dir = run_details['process_root']+'/run/stage_arguments/{}_{}'.format(run_details['network'],run_details['parameter'])

        #check if need to create any directories in path
        check_directory_existence(arguments_dir, directories_not_to_test='{}/run'.format(run_details['process_root']))

        #clean arguments directory 
        for zippath in glob.iglob(os.path.join(arguments_dir, '*.txt')):
            os.remove(zippath)
        for zippath in glob.iglob(os.path.join(arguments_dir, '*.grz')):
            os.remove(zippath)

        #-----------------------------#

        #define directory where jobs a from 
        submit_dir = run_details['process_root']+'/run/submit_stage/{}_{}'.format(run_details['network'],run_details['parameter'])

        #check if need to create any directories in path
        check_directory_existence(submit_dir, directories_not_to_test='{}/run'.format(run_details['process_root']))

        #clean submit directory 
        for zippath in glob.iglob(os.path.join(submit_dir, '*.sh')):
            os.remove(zippath)
        for zippath in glob.iglob(os.path.join(submit_dir, '*.rst')):
            os.remove(zippath)
        for zippath in glob.iglob(os.path.join(submit_dir, '*.log')):
            os.remove(zippath)


def write_station_specific_numpy(metadata,data,run_details,parameter_details,update_file=True):

    '''write out metadata/data to station specific numpy .npz file
       update_file variable is used to overwrite file when want to update it
    '''

    #get station reference
    station_reference = metadata['station_reference'][0]

    #get start and end of data ranges from file
    start_time = data['time'][0]
    end_time = data['time'][-1]+int(data['temporal_resolution'][-1])

    #deal with instances where temporary station file already exists and are not updating file - write dupN suffix variable for differentiation
    if not update_file:
        temporary_file_exists = os.path.isfile(run_details['data_root']+'{}/{}/native/{}/temporary_{}/{}_{}_{}.npz'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'],run_details['GHOST_stage'],station_reference,start_time,end_time))

        if temporary_file_exists:
            dup_n = 1
            while temporary_file_exists:
                if not os.path.isfile(run_details['data_root']+'{}/{}/native/{}/temporary_{}/{}_{}_{}:dup{}.npz'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'],run_details['GHOST_stage'],station_reference,start_time,end_time,dup_n)):
                    break
                dup_n +=1
            fname = run_details['data_root']+'{}/{}/native/{}/temporary_{}/{}_{}_{}:dup{}.npz'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'],run_details['GHOST_stage'],station_reference,start_time,end_time,dup_n)
            np.savez(fname,data=data,metadata=metadata)
            #give 770 permissions for file and make owner bsc32
            set_file_permissions_ownership(fname)
            return

    #write file for the first time or update it
    fname = run_details['data_root']+'{}/{}/native/{}/temporary_{}/{}_{}_{}.npz'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'],run_details['GHOST_stage'],station_reference,start_time,end_time)
    np.savez(fname,data=data,metadata=metadata)
    #give 770 permissions for file and make owner bsc32
    set_file_permissions_ownership(fname)
    return


def clean_native_directory(run_details, parameter_details):
        
    '''remove all existing station specific native resolution .npz files in native file directory for a specific network/parameter'''

    #set native directory for specific network/parameter
    native_dir = run_details['data_root'] + '{}/{}/native/{}/'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'])
    #iterate through all files in directory, removing each of them
    for zippath in glob.iglob(os.path.join(native_dir, '*.npz')):
        os.remove(zippath)


def clean_output_resolution_directories(run_details, parameter_details):
    
    '''remove all existing final output resolution netCDF files for a specific network/parameter, for all defined output resolutions'''

    from GHOST_standards import standard_temporal_resolutions

    #iterate through all defined temporal resolutions to process
    for temporal_resolution_code in run_details['temporal_resolutions']:
    
        #get temporal resolution specific variables
        temporal_resolution_vars = standard_temporal_resolutions[temporal_resolution_code]

        #iterate through all regions to process:
        for region in run_details['regions']:

            #set network directory per region
            if region == 'GLOBAL':
                network_dir = run_details['network']
            else:
                network_dir = '{}_{}'.format(run_details['network'], region)

            #set output resolution directory for specific network/output resolution/parameter
            output_resolution_dir = run_details['data_root'] + '{}/{}/{}/{}/'.format(network_dir,run_details['ghost_version'],temporal_resolution_vars['temporal_resolution_path'],parameter_details['bsc_parameter_name'])
            #iterate through all files in directory, removing each of them
            for zippath in glob.iglob(os.path.join(output_resolution_dir, '*.nc')):
                os.remove(zippath)


def get_data_files(run_details, parameter_details, specific_station='', specific_year='', specific_month='', specific_day=''):
    '''gets list of all relevant data files for network/parameter (and an optional specific station/year/month/day) combination in set date range'''

    from dateutil.relativedelta import relativedelta

    # create datetime objects of defined start date/end date.
    start_date_datetime = datetime.datetime(int(run_details['start_year']), int(run_details['start_month']), 1)
    end_date_datetime = datetime.datetime(int(run_details['end_year']), int(run_details['end_month']), 1)

    # create empty list to store valid data files
    valid_data_files = []

    if 'AERONET' in run_details['network']:
        #get data files from all different levels
        level_10_data_files = []
        level_15_data_files = glob.glob(run_details['data_root']+'{}/original_files/lev1.5/{}/*.html'.format(run_details['network'],parameter_details['aeronet_matrix']))
        level_20_data_files = []
        #merge data_files 
        data_files = level_10_data_files + level_15_data_files + level_20_data_files
        
        #create lists for storing all valid data file days/months/years
        valid_data_days = []
        valid_data_months = []
        valid_data_years = []
        
        # limit data files to just those with data inside defined data range
        for file_ii, data_file in enumerate(data_files):    
            file_data_day = data_file.split('/')[-1].replace('.html','')
            current_year = int(file_data_day[:4])
            current_month = int(file_data_day[4:6])
            current_day = int(file_data_day[6:8])
            start_date = datetime.datetime(current_year, current_month, current_day, 0, 0)
            end_date = datetime.datetime(current_year, current_month, current_day, 0, 0)  + datetime.timedelta(days=1)  
            if (start_date_datetime <= start_date < end_date_datetime) or (
                    start_date_datetime < end_date <= end_date_datetime) or (
                    start_date <= start_date_datetime < end_date) or (start_date < end_date_datetime <= end_date):
                valid_data_files.append(data_files[file_ii])
                valid_data_days.append(file_data_day)
                valid_data_months.append(file_data_day[:6])
                valid_data_years.append(file_data_day[:4])

    elif run_details['network'] == 'CANADA_NAPS':
        #handle complicated aerosol parameters
        #PM10, PM2.5, PM1
        if (run_details['parameter'] == 'PM10') or (run_details['parameter'] == 'PM2.5') or (run_details['parameter'] == 'PM1'):
            data_files = glob.glob(run_details['data_root']+'{}/original_files/{}/{}/mass/*.hly'.format(run_details['network'],parameter_details['matrix'],run_details['parameter']))
        #all other parameters
        else:
            data_files = glob.glob(run_details['data_root']+'{}/original_files/{}/{}/*.hly'.format(run_details['network'],parameter_details['matrix'],run_details['parameter']))

        #list to append valid data file years
        valid_data_years = []

        #limit data files to just those with data inside defined data range
        for data_file in data_files:
            data_year = int(data_file.split('/')[-1][:4])
            start_date = datetime.datetime(data_year, 1, 1, 0, 0)
            end_date = datetime.datetime(data_year+1, 1, 1, 0, 0)
            if (start_date_datetime <= start_date < end_date_datetime) or (
                    start_date_datetime < end_date <= end_date_datetime) or (
                    start_date <= start_date_datetime < end_date) or (start_date < end_date_datetime <= end_date):
                valid_data_files.append(data_file)
                valid_data_years.append(str(data_year))

    elif run_details['network'] == 'CAPMoN':
        data_files = glob.glob(run_details['data_root']+'{}/original_files/{}/{}/*.csv'.format(run_details['network'],parameter_details['matrix'],run_details['parameter']))
        
        #list to append valid data file years
        valid_data_years = []
        
        # limit data files to just those with data inside defined data range
        for data_file in data_files:
            data_year = int(data_file.split('/')[-1].split('.csv')[-1][-4:])
            start_date = datetime.datetime(data_year, 1, 1, 0, 0)
            end_date = datetime.datetime(data_year+1, 1, 1, 0, 0)
            if (start_date_datetime <= start_date < end_date_datetime) or (
                    start_date_datetime < end_date <= end_date_datetime) or (
                    start_date <= start_date_datetime < end_date) or (start_date < end_date_datetime <= end_date):
                valid_data_files.append(data_file)
                valid_data_years.append(str(data_year))

    elif run_details['network'] == 'EBAS':

        # PM10, PM2.5, PM1
        if run_details['parameter'] in ['PM10','PM2.5','PM1']:
            data_files = glob.glob(
                run_details['data_root'] + '{}/original_files/aerosol/{}/mass/*.nas'.format(run_details['network'],
                                                                                            run_details['parameter']))           
        # bulk speciated
        elif parameter_details['matrix'] == 'aerosol':
            data_files = glob.glob(
                run_details['data_root'] + '{}/original_files/aerosol/bulk/speciated/{}/*.nas'.format(
                    run_details['network'], run_details['parameter']))
            #also extend data files list with corrected component files (C, OC)
            if run_details['parameter'] in ['C','OC']:
                data_files.extend(glob.glob(run_details['data_root'] + '{}/original_files/aerosol/bulk/speciated/{}_corrected/*.nas'.format(
                    run_details['network'], run_details['parameter'])))
        # PM10/PM2.5/PM1 speciated
        elif parameter_details['matrix'] in ['pm10','pm2.5','pm1']:
            aerosol_type = run_details['parameter'].split('_')[0]
            param = run_details['parameter'].split('_')[-1]
            data_files = glob.glob(run_details['data_root'] + '{}/original_files/aerosol/{}/speciated/{}/*.nas'.format(
                run_details['network'], aerosol_type, param)) 
            #also extend data files list with corrected component files (C, OC)
            if run_details['parameter'] in ['PM10_C','PM10_OC','PM2.5_C','PM2.5_OC','PM1_C','PM1_OC']:
                data_files.extend(glob.glob(run_details['data_root'] + '{}/original_files/aerosol/{}/speciated/{}_corrected/*.nas'.format(
                run_details['network'], aerosol_type, param))) 
        # all other parameters
        else:
            data_files = glob.glob(
                run_details['data_root'] + '{}/original_files/{}/{}/*.nas'.format(run_details['network'],
                                                                                  parameter_details['matrix'],
                                                                                  run_details['parameter']))

        # get start and end timepoints of data files
        data_files_without_root = [data_file.split('/')[-1] for data_file in data_files]
        data_files_start_dates = [data_file.split('.')[1] for data_file in data_files_without_root]
        data_files_start_dates = [
            datetime.datetime(int(start_date[:4]), int(start_date[4:6]), int(start_date[6:8]), int(start_date[8:10]),
                              int(start_date[10:12])) for start_date in data_files_start_dates]
        data_files_file_lengths = [data_file.split('.')[6] for data_file in data_files_without_root]
        for file_ii, start_date in enumerate(data_files_start_dates):
            file_length_int = int(re.findall(r'\d+', data_files_file_lengths[file_ii])[0])
            file_length_str = re.findall(r'\D+', data_files_file_lengths[file_ii])[0]
            if file_length_str == 'h':
                end_date = start_date + relativedelta(hours=file_length_int)
            elif file_length_str == 'd':
                end_date = start_date + relativedelta(days=file_length_int)
            elif file_length_str == 'w':
                end_date = start_date + relativedelta(weeks=file_length_int)
            elif file_length_str == 'mo':
                end_date = start_date + relativedelta(months=file_length_int)
            elif file_length_str == 'y':
                end_date = start_date + relativedelta(years=file_length_int)

            # limit data files to just those with data inside defined data range
            if (start_date_datetime <= start_date < end_date_datetime) or (
                    start_date_datetime < end_date <= end_date_datetime) or (
                    start_date <= start_date_datetime < end_date) or (start_date < end_date_datetime <= end_date):
                valid_data_files.append(data_files[file_ii])

        #get station references from all valid data files
        file_station_references = [f.split('/')[-1].split('.')[0] for f in valid_data_files] 

    elif 'EEA_AQ_eReporting' in run_details['network']:
        # gather data files from both E1a/E2a streams

        # PM10, PM2.5, PM1
        if run_details['parameter'] in ['PM10','PM2.5','PM1']:
            stem_E1a = run_details['data_root'] + '{}/original_files/E1a/aerosol/{}/mass'.format(run_details['network'],run_details['parameter'])
            stem_E2a = run_details['data_root'] + '{}/original_files/E2a/aerosol/{}/mass'.format(run_details['network'],run_details['parameter'])
        # bulk speciated
        elif parameter_details['matrix'] == 'aerosol':
            stem_E1a = run_details['data_root'] + '{}/original_files/E1a/aerosol/bulk/speciated/{}'.format(run_details['network'], run_details['parameter'])
            stem_E2a = run_details['data_root'] + '{}/original_files/E2a/aerosol/bulk/speciated/{}'.format(run_details['network'], run_details['parameter'])        
        # PM10/PM2.5/PM1 speciated
        elif parameter_details['matrix'] in ['pm10','pm2.5','pm1']:
            aerosol_type = run_details['parameter'].split('_')[0]
            param = run_details['parameter'].split('_')[-1]
            stem_E1a = run_details['data_root'] + '{}/original_files/E1a/aerosol/{}/speciated/{}'.format(run_details['network'], aerosol_type, param)
            stem_E2a = run_details['data_root'] + '{}/original_files/E2a/aerosol/{}/speciated/{}'.format(run_details['network'], aerosol_type, param)
        # all other parameters
        else:
            stem_E1a = run_details['data_root'] + '{}/original_files/E1a/{}/{}'.format(run_details['network'], parameter_details['matrix'], run_details['parameter'])
            stem_E2a = run_details['data_root'] + '{}/original_files/E2a/{}/{}'.format(run_details['network'], parameter_details['matrix'], run_details['parameter'])
                
        # gather all data files
        data_files = []
        for stem in [stem_E1a, stem_E2a]:
            for root, dirs, files in os.walk(stem):
                for f in files:
                    if f.endswith("timeseries.csv"):
                        data_files.append(os.path.join(root, f))
                
        # get start and end timepoints of data files
        data_files_start_dates = [datetime.datetime(int(data_file.split('/')[-1].split('_')[3]), 1, 1) for
                                  data_file in data_files]
        for file_ii, start_date in enumerate(data_files_start_dates):
            end_date = start_date + relativedelta(years=1)

            # limit data files to just those with data inside defined data range
            if (start_date_datetime <= start_date < end_date_datetime) or (
                    start_date_datetime < end_date <= end_date_datetime) or (
                    start_date <= start_date_datetime < end_date) or (
                    start_date < end_date_datetime <= end_date):
                valid_data_files.append(data_files[file_ii])

        #get station references from all valid data files
        f_split = [f.split('/')[-1].split('_') for f in valid_data_files]
        file_station_references = ['{}_{}_{}'.format(spl[0],spl[1],spl[2]) for spl in f_split]

    elif 'NCDC_ISD' in run_details['network']:
        data_files = []
        stem = run_details['data_root'] + '{}/original_files/meteo'.format(run_details['network'])
        for root, dirs, files in os.walk(stem):
            for f in files:
                data_files.append(os.path.join(root, f))

        file_station_references = []
        for file in data_files:
            stationref = str(file).split('/')[-1][:6] + str(file).split('/')[-1][7:12]
            stationyear = str(file).split('/')[-1][-4:]

            start_date = datetime.datetime(int(stationyear),1,1,0,0)
            end_date = datetime.datetime(int(stationyear)+1,1,1,0,0)

            if (start_date_datetime <= start_date < end_date_datetime) or (
                    start_date_datetime < end_date <= end_date_datetime) or (
                    start_date <= start_date_datetime < end_date) or (
                    start_date < end_date_datetime <= end_date):
                valid_data_files.append(file)
                file_station_references.append(stationref)

    # if a specific station variable has been defined, limit data files to be just those associated with the defined station reference
    if specific_station != '':
        valid_data_files = [data_file for data_file, file_station_reference in
                            zip(valid_data_files, file_station_references) if
                            str(specific_station) == str(file_station_reference)]
    # if a specific data year variable has been defined, limit data files to be just those associated with the defined data year
    if specific_year != '':
        valid_data_files = [data_file for data_file, file_data_year in zip(valid_data_files, valid_data_years) if
                            str(specific_year) == str(file_data_year)]

    # if a specific data month variable has been defined, limit data files to be just those associated with the defined data month
    if specific_month != '':
        valid_data_files = [data_file for data_file, file_data_month in zip(valid_data_files, valid_data_months) if
                            str(specific_month) == str(file_data_month)]

    # if a specific data day variable has been defined, limit data files to be just those associated with the defined data day  
    if specific_day != '':
        valid_data_files = [data_file for data_file, file_data_day in zip(valid_data_files, valid_data_days) if
                            str(specific_day) == str(file_data_day)]


    return sorted(valid_data_files)


def get_YYYY_from_filenames(data_files, run_details):

    '''get unique list of data years as YYYY strings to process from data filenames.
       process for this differs by network.
    ''' 

    data_years = []

    if 'AERONET' in run_details['network']:
        for i, f in enumerate(data_files):
            data_years.append(f.split('/')[-1][:4])
        data_years = np.unique(data_years)

    return data_years

def get_YYYYMM_from_filenames(data_files, run_details):

    '''get unique list of data months as YYYYMM strings to process from data filenames.
       process for this differs by network.  
    ''' 

    data_months = []
    
    if 'AERONET' in run_details['network']:
        for i, f in enumerate(data_files):
            data_months.append(f.split('/')[-1][:6])
        data_months = np.unique(data_months)
    
    return data_months

def get_YYYYMMDD_from_filenames(data_files, run_details):

    '''get unique list of data days as YYYYMMDD strings to process from data filenames.
       process for this differs by network.  
    ''' 

    data_days = []
    
    if 'AERONET' in run_details['network']:
        for i, f in enumerate(data_files):
            data_days.append(f.split('/')[-1].split('.html')[0])
        data_days = np.unique(data_days)
    
    return data_days

def get_station_references_from_filenames(data_files, run_details):

    '''get unique list of raw station references provided in data filenames.
       process for this differs by network.
    '''

    station_references = []

    if run_details['network'] == 'EBAS':
        for i, f in enumerate(data_files):
            f = f.split('/')[-1]
            station_references.append(f.split('.')[0])

    if 'EEA_AQ_eReporting' in run_details['network']:
        for i, f in enumerate(data_files):
            f_split = f.split('/')[-1]
            f_split = f_split.split('_')

            station_references.append('{}_{}_{}'.format(f_split[0],f_split[1],f_split[2]))

    if 'NCDC_ISD' in run_details['network']:
        for f in enumerate(data_files):
            f_split = f[-1][-17:].split('-')

            station_references.append('{}{}'.format(f_split[0], f_split[1]))

    #sort station references in a human-intuitive order
    station_references = sorted(remove_duplicates_preserve_order(station_references), key=natural_key)

    return station_references


def determine_key_changing_metadata(old_metadata, new_metadata):

    '''function for determining if certain key metadata fields have changed between files
       returns True if metadata has changed, False if not
    '''

    #define key metadata fields to check
    key_metadata_fields = ['latitude', 'longitude', 'altitude', 'sampling_height', 'measurement_altitude', 'distance_to_building', 'distance_to_kerb', 'distance_to_junction', 'distance_to_source', 'street_width', 'street_type', 'daytime_traffic_speed', 'daily_passing_vehicles', 'horizontal_geodetic_datum', 'data_level', 'climatology', 'station_name', 'city', 'country', 'population', 'representative_radius', 'associated_networks',
                           'standardised_network_provided_area_classification', 'standardised_network_provided_station_classification', 'standardised_network_provided_main_emission_source', 'standardised_network_provided_land_use', 'standardised_network_provided_terrain', 'standardised_network_provided_measurement_scale',
                           'primary_sampling_type', 'primary_sampling_instrument_name', 'primary_sampling_instrument_reported_flow_rate',
                           'sample_preparation_types', 'sample_preparation_techniques', 
                           'measurement_methodology', 'measuring_instrument_name', 'measuring_instrument_sampling_type', 'measuring_instrument_reported_flow_rate', 'measuring_instrument_reported_lower_limit_of_detection', 'measuring_instrument_reported_upper_limit_of_detection', 'measuring_instrument_reported_uncertainty', 'measuring_instrument_reported_accuracy', 'measuring_instrument_reported_precision', 'measuring_instrument_reported_measurement_resolution', 'measuring_instrument_reported_absorption_cross_section', 'measuring_instrument_calibration_scale',
                           'network_provided_volume_standard_temperature','network_provided_volume_standard_pressure']

    #iterate through key metadata fields and check to see if there are differences between old/new metadata
    for key_metadata_field in key_metadata_fields:
        old_value = old_metadata[key_metadata_field]
        new_value = new_metadata[key_metadata_field]

        #if both values are NaN, continue to next field
        if (pd.isnull(old_value)) & (pd.isnull(new_value)):
            continue

        #test equivalence, if not equal, break out of function, as already know key metadata not equal between files
        if old_value != new_value:
            return True

    #if have made it through iteration of key metadata fields, key metadata is therefore equivalent
    return False


def redo_standardisations(dataframe, parameter_details, fields_to_standardise=None):

    '''take numpy dataframe for metadata/data and a list of fields that have been updated/modified ensure are standardised correctly.
       this is imperative for string formatting, e.g. to ensure capitalisations are correct, but the function will check standards of all passed fields.
    '''

    from GHOST_standards import get_standard_data, get_standard_metadata

    #get standard formats for dataframe
    if 'latitude' in dataframe.dtype.names:
        standard_dict = get_standard_metadata(parameter_details)
    else:
        standard_dict = get_standard_data(parameter_details)

    #define fields to standardise to be all fields if None defined
    if fields_to_standardise is None:
        fields_to_standardise = dataframe.dtype.names        

    #iterate through provided fields
    for key in fields_to_standardise:

        if key != 'measurement_qa_flags':

            #get list of values for field
            param_array = dataframe[key]

            #get standard data type for field
            standard_dtype = dataframe[key].dtype

            #get string format for field
            string_format = standard_dict[key]['string_format']

            #do string formatting for relevant fields before filling dataframe
            if (standard_dtype == np.dtype('O')) & (key not in ['flag', 'qa']):

                #upper case string formatting
                if 'upper' in string_format:
                    param_array = [np.str(x).upper() if pd.isnull(x) == False else x for x in param_array]

                #lower case string formatting
                elif 'lower' in string_format:
                    param_array = [np.str(x).lower() if pd.isnull(x) == False else x for x in param_array]

                #titalling string formatting
                elif 'title' in string_format:
                    param_array = [np.str(x).title() if pd.isnull(x) == False else x for x in param_array]

                #standard string formatting
                else:
                    param_array = [np.str(x) if pd.isnull(x) == False else x for x in param_array]

            #rewrite dataframe
            dataframe[key] = param_array

    return dataframe


def drop_stations(station_references, max_N_stations):

    '''randomly all drop stations > than a particular max N stations number.'''

    import random

    #if N station references is <= max_stations, then return
    if len(station_references) <= max_N_stations:
        return station_references

    #shuffle station references randomly
    random.shuffle(station_references)

    #keep first max_N_stations station references preferentially
    return station_references[:max_N_stations] 


def list_duplicates(seq):

    '''Function for finding and returning values/indices of all duplicates in a string/list
       Returns list of tuples.
       The first value in each tuple is the duplicated value
       The second value in each tuple is a list of duplicated indices
       Returns empty list if no duplicates are found
       Function is optimised for speed
    '''

    from collections import defaultdict

    tally = defaultdict(list)
    for i,item in enumerate(seq):
        tally[item].append(i)
    return sorted(((key,locs) for key,locs in tally.items()
                            if len(locs)>1))


def remove_duplicates_preserve_order(seq):

    '''remove duplicates from list while preserving order'''

    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


def merge_lists_sharing_common_elements(l):

    '''taking a list of lists merge all the lists that share common elements
       source: https://stackoverflow.com/questions/4842613/merge-lists-that-share-common-elements/4843408#4843408
    '''

    l = copy.deepcopy(l)
    len_l = len(l)
    i = 0
    while i < (len_l - 1):
        for j in range(i + 1, len_l):
            # i,j iterate over all pairs of l's elements including new
            # elements from merged pairs. We use len_l because len(l)
            # may change as we iterate
            i_set = set(l[i])
            j_set = set(l[j])
            if len(i_set.intersection(j_set)) > 0:
                # Remove these two from list
                l.pop(j)
                l.pop(i)
                # Merge them and append to the orig. list
                ij_union = list(i_set.union(j_set))
                l.append(ij_union)
                # len(l) has changed
                len_l -= 1
                # adjust 'i' because elements shifted
                i -= 1
                # abort inner loop, continue with next l[i]
                break
        i += 1
    return l


def natural_key(string_):

    '''key for sorting in human natural order'''

    return [int(s) if s.isdigit() else s for s in re.split(r'(\d+)', string_)]


def read_from_numpy(numpy_file, desired_variables):

    '''read numpy .npz files, returning desired variables (input as list) stored in a dictionary: {desired_var_name: var}'''

    numpy_dataframe = np.load(numpy_file, allow_pickle=True)

    loaded_variables = {}

    for desired_variable in desired_variables:
        loaded_variables[desired_variable] = numpy_dataframe[desired_variable]

    #close .npz file
    numpy_dataframe.close()

    return loaded_variables


def cut_data_array(data_array,data_type,period_start,period_end):

    '''function which cuts data/metadata arrays by a given data range, provided as start minute from reference + end minute from reference'''

    if data_type == 'data':
        #cut relevant indices
        start_i = bisect.bisect_left(data_array['time'], period_start)
        end_i =   bisect.bisect_left(data_array['time'], period_end)

        #cut data array
        data_array = data_array[start_i:end_i]

    if data_type == 'metadata':

        #cut relevant indices
        start_i = bisect.bisect_left(data_array['meta_update_stamp'], period_start)
        end_i =   bisect.bisect_left(data_array['meta_update_stamp'], period_end)

        #flag for determining if need to update first metadata update stamp
        update_start_stamp = False

        #if only 1 metadata value, and data period starts after when metadata start stamp, then modify start_meta_i to be 0
        #also set flag for updating of meta update stamp
        if len(data_array['meta_update_stamp']) == 1:
            if period_start > data_array['meta_update_stamp']:
                start_i = 0
                update_start_stamp = True

        #if have more than 1 metadata value and data range starts between metadata update periods, then adjust first metadata field to be from previous index
        if len(data_array['meta_update_stamp']) > 1: 
            if (period_start < data_array[start_i]['meta_update_stamp']) & (start_i != 0):
                start_i = start_i - 1
                update_start_stamp = True

        #cut metadata array
        data_array = data_array[start_i:end_i]

        #if need to update first metadata time stamp, do so now
        if update_start_stamp:
            data_array[0]['meta_update_stamp'] = period_start

    return data_array


def get_unique_processed_references(run_details, parameter_details):

    '''get unique processed station references from files saved out in the native/temporary file directories based on GHOST stage you are currently in'''

    processed_references = []

    #if current GHOST stage is in S stages, gather station files from (GHOST_stage-1)_temporary directory
    if 'S' in run_details['GHOST_stage']:
        previous_GHOST_stage = 'S{}'.format(int(run_details['GHOST_stage'][1:])-1)
        station_files = glob.glob(run_details['data_root']+'{}/{}/native/{}/temporary_{}/*.npz'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'],previous_GHOST_stage))
    #otherwise gather finalised station files from native directory
    else:
         station_files = glob.glob(run_details['data_root']+'{}/{}/native/{}/*.npz'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name']))

    for file_i, file in enumerate(station_files):
        split = file.split('/')[-1].split('_')
        station_reference = '_'.join(split[:-2])
        processed_references.append(station_reference)

    #sort references in alphabetical order
    processed_references = sorted(np.unique(processed_references), key=natural_key)

    return processed_references


def get_data_from_reference(station_reference,run_details,parameter_details,read='all'):

    '''take station reference and return loaded metadata/data
       the type of data returned ('read') can be configured. By default both final stage metadata and data read and returned.
       read = 'all' --> read and return data + metadata (and filename); read = 'data' --> just read and return data (and filename); read = 'metadata' --> just read and return metadata (and filename)
    '''

    #if GHOST stage is in S stages, gather station filename from (GHOST_stage-1)_temporary directory
    if 'S' in run_details['GHOST_stage']:
        previous_GHOST_stage = 'S{}'.format(int(run_details['GHOST_stage'][1:])-1)
        station_numpy_filename = glob.glob(run_details['data_root']+'{}/{}/native/{}/temporary_{}/{}_*'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'],previous_GHOST_stage,station_reference))[0]
     #otherwise gather filename from native directory
    else:
        station_numpy_filename = glob.glob(run_details['data_root']+'{}/{}/native/{}/{}_*'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'],station_reference))[0]

    #set desired variables to read
    if read == 'metadata':
        desired_variables = ['metadata']
    elif read == 'data':
        desired_variables = ['data']
    elif read == 'all':
        desired_variables = ['metadata', 'data']

    #read variables
    loaded_variables = read_from_numpy(station_numpy_filename, desired_variables)

    #return variables
    if read == 'metadata':
        return loaded_variables['metadata'], station_numpy_filename
    elif read == 'data':
        return loaded_variables['data'], station_numpy_filename
    elif read == 'all':
        return loaded_variables['metadata'], loaded_variables['data'], station_numpy_filename


def normalise_by_minmax(values, lower_limit=0.0, upper_limit=1.0, reverse=False):
    
    '''function that normalises list of values to lie within a given range (by default 0 to 1)
       the minimum value of the array will lie on the lower_limit,
       and the maximum_value of the array will lie on the upper_limit
       if reverse if set to be true, then the minimum value will lie on the upper limit,
       and the maximum value will lie on the lower limit.
    '''

    values = np.array(values, dtype=np.float32)
    uniques = np.unique(values)
    if len(uniques) == 0:
        return values
    elif len(uniques) == 1:
        return np.array([upper_limit]*len(values), dtype=np.float32)
    if reverse:
        values = values * -1.0
    width = upper_limit - lower_limit
    return (values - values.min())/(values.max() - values.min()) * width + lower_limit


def normalise_by_max(values, reverse=False, offset=0.0):
    
    '''function that normalises list of values by the maximum value in the array
       the maximum value will be normalised to be 1, and the rest as a ratio to the max
       if reverse is set to be True, then the ratio is made to the maximum after inverting the array
       if offset is non-zero, then the offset value is subsequently applied to all values.
       the input list of values input must be non-negative
    '''

    values = np.array(values, dtype=np.float32)
    uniques = np.unique(values)
    if len(uniques) == 0:
        return values
    elif len(uniques) == 1:
        return np.array([1.0+offset]*len(values), dtype=np.float32)

    if reverse:
        previous_min = np.min(values)
        values = values * -1.0
        values = values + (previous_min - (np.min(values)))
    return (values/np.max(values))+offset


def normalise_by_defined_max(values, max_val=None, offset=0.0):

    '''function that normalises list of values by a defined maximum value (if not set, by default will be max in data array)
       all values in the supplied data arrat will be set as a ratio to the max
       if offset is non-zero, then the offset value is subsequently applied to all values.
       the input list of values input must be non-negative
    '''

    values = np.array(values, dtype=np.float32)

    if max_val is None:
        max_val = np.max(values)   

    uniques = np.unique(values)
    if len(uniques) == 0:
        return values

    return (values/max_val)+offset


def return_PDF_intersection(pdf_1, pdf_2):
    
    '''function which returns intersection (between 0.0 and 1.0) between 2 PDFs with equivalent sample bases'''

    minima = np.minimum(pdf_1, pdf_2)
    intersection = np.true_divide(np.sum(minima), np.sum(pdf_2))
    return intersection


def filter_by_qa(qa_to_filter_by, data, qa, return_valid_mask=False):

    '''function that filters measurement data by a list of QA flags'''

    invalid_qa_flag_codes = [standard_QA_name_to_QA_code[invalid_flag_name] for invalid_flag_name in qa_to_filter_by]
    valid_mask = np.array([all(x not in qa_flag_codes for x in invalid_qa_flag_codes) for qa_flag_codes in qa])
    
    if return_valid_mask == False:
        return data[valid_mask]
    else:
        return valid_mask
    

def weighted_mode(list_of_values, weights):

    '''taking a list of values, and a list of weights associated with the values, calculate the weighted modal value
       this simply works by calculating the unique values in the list, then separately summing the weights associated with each unique value, and then determining which unique value is associated with the highest cumulative weights.
       if there is a tie, the lowest sorted value will be kept preferentially (i.e. a instead of b, 1 instead of 2)
       if there is only 1 unique value, then modal value is simply the only unique value
    '''

    #get unique values in list
    unique_values = np.unique(list_of_values)

    #if only 1 unique value, then modal value is simply the only unique value, return function
    if len(unique_values) == 1:
        return unique_values[0]

    #add all unique values to weight counter dictionary as keys, and give an initial value of 0.0
    weight_counter = {}
    for unique_value in unique_values:
        weight_counter[unique_value] = 0.0

    #iterate through all values in the list of values and associated weights
    for value, weight in zip(list_of_values,weights):
        #add weight associated with value to weight counter dictionary
        weight_counter[value] += weight

    #the unique value key associated the the greatest sum of weights in the weight counter dictionary is the modal value
    #if there is a tie, the lowest sorted key will be kept preferentially (i.e. a instead of b, 1 instead of 2)
    modal_value = max(weight_counter.keys(), key=(lambda key: weight_counter[key]))

    return modal_value


def get_relevant_yearmonths(run_details, parameter_details, temporal_resolution_code, region):

    '''create list of yearmonth strings, for which in those time periods exists valid processed station data''' 

    from GHOST_standards import standard_temporal_resolutions

    #get output temporal resolution specific variables
    temporal_resolution_vars = standard_temporal_resolutions[temporal_resolution_code]

    #set network directory where to read data from
    if region == 'GLOBAL':
        network_read_dir = run_details['network']
    else:
        network_read_dir = '{}_{}'.format(run_details['network'], region)

    #create list for storing relevant yearmonth strings
    yearmonth_strings = []

    #create list of all years data has been processed for
    years = np.arange(int(run_details['start_year']),int(run_details['end_year'])+1)

    #iterate through years
    for year in years:

        #is there less than 1 year of data?
        if len(years) == 1:
            months = np.arange(int(run_details['start_month']), int(run_details['end_month']))

        #else, is it first process year?
        elif year == years[0]:
            months = np.arange(int(run_details['start_month']), 13)

        #else, is it last process year?
        elif year == years[-1]:
            months = np.arange(1,int(run_details['end_month']))

        #full year of months otherwise
        else:
            months = np.arange(1,13)

        #iterate through months
        for month in months:

            if month < 10:
                month_string = '0'+str(month)
                yearmonth_str = str(year)+month_string
            else:
                month_string = str(month)
                yearmonth_str = str(year)+month_string

            #check if have station files in year/month directory
            year_month_files = sorted(glob.glob(run_details['data_root'] + '{}/{}/{}/{}/temporary/{}/{}/*.nc'.format(network_read_dir,run_details['ghost_version'],temporal_resolution_vars['temporal_resolution_path'],parameter_details['bsc_parameter_name'],year,month_string)))

            #if so, then append yearmonth_str to list
            if len(year_month_files) > 0:
                yearmonth_strings.append(yearmonth_str)

    return yearmonth_strings